/*
 * Common image-related functions.
 *
 * Copyright (C) 2018 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __IMAGE_H__
#define __IMAGE_H__

#include <stdint.h>



/* Image data type. */

typedef struct
{
    /* Image width and height. */

    uint32_t width, height;

    /* Image data. */

    uint8_t image[];

} image_t;



/* Stored display regions. */

typedef struct
{
    /* Position of the stored region on the screen. */

    int x, y;

} position_t;

typedef struct
{
    /* Stored region overplotted by the sprite. */

    uint8_t *image;
    uint32_t size;

    /* Number of stored frames. */

    int stored;

} stored_regions_t;



/* A simple sprite abstraction. */

typedef struct
{
    /* The sprite image. */

    image_t *image;

    /* The stored regions to be used to unplot the sprite. */

    stored_regions_t *regions;

    /* Sprite frame positions. */

    position_t *pos;

    /* Colour key for transparency. */

    int key;

    /* The display associated with the sprite. */

    display_config_t *cfg;

} sprite_t;



/* Initialise stored regions, allocating memory and a structure to access it.

   Stored_Regions(<name>, int frames, int size)
*/

#define Stored_Regions(NAME, FRAMES, SIZE)          \
    uint8_t __##NAME##_image[(FRAMES) * (SIZE)];    \
    stored_regions_t NAME = {                       \
        .image = __##NAME##_image,                  \
        .size = SIZE,                               \
        .stored = 0};

/* Access functions. */

position_t *image_get_stored_position(sprite_t *s, int frame);

uint8_t *image_get_stored_region(stored_regions_t *r, int frame);



/* Initialise a sprite object using an existing image without the capability of
   storing regions of the screen.

   SpriteOverwriting(<name>, image_t *image, display_config_t *cfg, int frames,
                     int key)
*/

#define SpriteOverwriting(NAME, IMAGE, CFG, FRAMES, KEY)    \
    position_t __##NAME##_pos[FRAMES];                      \
                                                            \
    sprite_t NAME = {                                       \
        .image = IMAGE,                                     \
        .regions = 0,                                       \
        .key = KEY,                                         \
        .cfg = CFG,                                         \
        .pos = __##NAME##_pos};

/* Initialise a sprite object using an existing image, creating stored regions
   for the animation of the sprite.

   Sprite(<name>, image_t *image, display_config_t *cfg, int frames, int key)
*/

#define Sprite(NAME, IMAGE, CFG, FRAMES, KEY)           \
    Stored_Regions(__##NAME##_regions, FRAMES,          \
                   (IMAGE)->width * (IMAGE)->height);   \
    position_t __##NAME##_pos[FRAMES];                  \
                                                        \
    sprite_t NAME = {                                   \
        .image = IMAGE,                                 \
        .regions = &(__##NAME##_regions),               \
        .key = KEY,                                     \
        .cfg = CFG,                                     \
        .pos = __##NAME##_pos};

/* Sprite plotting operations. */

void image_plot_sprite(sprite_t *s);

void image_unplot_sprite(sprite_t *s);

void image_unplot_sprite_from_image(sprite_t *s, sprite_t *bg,
                                    int xorigin, int yorigin);

/* Update operations. */

position_t *image_get_sprite_position(sprite_t *s);

void image_set_sprite_position(sprite_t *s, int x, int y);

/* Convenience operations. */

void image_plot_sprite_section(sprite_t *s,
                               int xstart, int ystart, int xsize, int ysize,
                               int x, int y, int key);

void image_tile_sprite(sprite_t *s, int xsource, int ysource,
                                    int width, int height,
                                    int xdisplay, int ydisplay);

void image_update_scrolled_tiled_image(sprite_t *s, int xorigin, int yorigin,
                                                    int xstep, int ystep);

void image_update_tiled_image(sprite_t *s, int xorigin, int yorigin,
                                           int width, int height,
                                           int xdisplay, int ydisplay);

#endif /* __IMAGE_H__ */
