/*
 * Common image-related functions.
 *
 * Copyright (C) 2018 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "display.h"
#include "image.h"
#include "utils.h"



/* Obtain the position for the stored region from the given frame. */

position_t *image_get_stored_position(sprite_t *s, int frame)
{
    return &s->pos[frame];
}

/* Obtain the image data for the stored region from the given frame. */

uint8_t *image_get_stored_region(stored_regions_t *r, int frame)
{
    return r->image + r->size * frame;
}



/* Get a sprite's position. */

position_t *image_get_sprite_position(sprite_t *s)
{
    return image_get_stored_position(s, s->cfg->frame);
}

/* Set a sprite's position. */

void image_set_sprite_position(sprite_t *s, int x, int y)
{
    position_t *p = image_get_sprite_position(s);

    p->x = x;
    p->y = y;
}



/* Copy a region from the screen to the store, then blit the image. */

void image_plot_sprite(sprite_t *s)
{
    int frame = 0;
    position_t *p = 0;

    if (s->regions)
    {
        frame = s->cfg->frame;
        p = image_get_stored_position(s, frame);

        /* Copy to the stored region. */

        display_copy(s->cfg, image_get_stored_region(s->regions, frame),
                     s->image->width, s->image->height,
                     p->x, p->y, -1, 0);
    }

    /* Plot to the screen. */

    display_copy(s->cfg, s->image->image,
                 s->image->width, s->image->height,
                 p->x, p->y, s->key, 1);

    if (s->regions)
    {
        /* Record the stored background details. */

        if (frame >= s->regions->stored)
            s->regions->stored = frame + 1;
    }
}

/* Copy a region from the store to the screen, restoring the original
   background. */

void image_unplot_sprite(sprite_t *s)
{
    int frame;
    position_t *p;

    if (!s->regions)
        return;

    frame = s->cfg->frame;
    p = image_get_stored_position(s, frame);

    /* Only unplot the sprite if a region was stored for the frame. */

    if (s->regions->stored > frame)
        display_copy(s->cfg, image_get_stored_region(s->regions, frame),
                     s->image->width, s->image->height,
                     p->x, p->y, -1, 1);
}

/* Unplot a sprite by restoring a region from the background image. */

void image_unplot_sprite_from_image(sprite_t *s, sprite_t *bg,
                                    int xorigin, int yorigin)
{
    int frame = s->cfg->frame;
    position_t *p = image_get_stored_position(s, frame);

    /* Plot the region of the background using the sprite image dimensions
       converted to background image dimensions at the sprite's position on the
       display. */

    image_update_tiled_image(bg, xorigin, yorigin,
                             s->image->width, s->image->height,
                             p->x, p->y);
}



/* Plot a section of an image without storing the background beforehand. */

void image_plot_sprite_section(sprite_t *s,
                               int xstart, int ystart, int xsize, int ysize,
                               int x, int y, int key)
{
    display_copy_section(s->cfg, s->image->image,
                         s->image->width, s->image->height,
                         xstart, ystart, xsize, ysize,
                         x, y, key, 1);
}



/* Tile a sprite, using the given source origin, filling a display region. */

void image_tile_sprite(sprite_t *s, int xsource, int ysource,
                                    int width, int height,
                                    int xdisplay, int ydisplay)
{
    /* Determine the portion of the sprite to be plotted in the first column. */

    int source_width = s->image->width - xsource;
    int source_height;
    int total_height;
    int x, y;
    int xs, ys;

    /* Fill (xdisplay, ydisplay) with (width, height) from source, slice by
       slice. */

    x = xdisplay;
    xs = xsource;

    while (width)
    {
        /* Fill (x, ydisplay) with (source_width, height) from source, with the
           height being divided into image-sized pieces. */

        total_height = height;
        source_height = s->image->height - ysource;
        y = ydisplay;
        ys = ysource;

        while (total_height)
        {
            /* Plot as much of the image as is available from the given source
               coordinates. */

            image_plot_sprite_section(s, xs, ys,
                min(width, source_width), min(total_height, source_height),
                x, y, -1);

            /* Continue to plot the image again to fill the slice. */

            if (source_height >= total_height)
                break;

            total_height -= source_height;
            y += source_height;
            ys = 0;
            source_height = s->image->height;
        }

        /* Get the next slice of the column. */

        if (source_width >= width)
            break;

        width -= source_width;
        x += source_width;
        xs = 0;
        source_width = s->image->width;
    }
}

/* Plot a scrolling tiled image upon a viewport update. */

void image_update_scrolled_tiled_image(sprite_t *s, int xorigin, int yorigin,
                                                    int xstep, int ystep)
{
    /* The display regions are either the left or right edge... */

    int xedge = xstep < 0 ? 0 : s->cfg->line_length - xstep;
    int xdisplay = xedge;

    /* and either the top or bottom edge... */

    int yedge = ystep < 0 ? 0 : s->cfg->line_count - ystep;
    int ydisplay = yedge;

    /* Determine the origin position within the image. */

    int xpos = wrap_value(xorigin, s->image->width);
    int ypos = wrap_value(yorigin, s->image->height);
    int xsource, ysource;

    /* Horizontal scrolling requires columns spanning the height of the screen
       at the appropriate edge (left or right). */

    /* The column width is the absolute increment. */

    if (xstep)
    {
        /* Find the source position for the appropriate edge. */

        xsource = wrap_value(xpos + xedge, s->image->width);
        ysource = ypos;

        /* Request tiling in the source coordinates. */

        image_tile_sprite(s, xsource, ysource,
                          abs(xstep), s->cfg->line_count,
                          xdisplay, 0);
    }

    /* Vertical scrolling requires columns across the width of the screen at the
       appropriate edge (top or bottom). */

    if (ystep)
    {
        /* Find the source position for the appropriate edge. */

        xsource = xpos;
        ysource = wrap_value(ypos + yedge, s->image->height);

        /* Request tiling in the source coordinates. */

        image_tile_sprite(s, xsource, ysource,
                          s->cfg->line_length, abs(ystep),
                          0, ydisplay);
    }
}

/* Plot a region of a tiled image. */

void image_update_tiled_image(sprite_t *s, int xorigin, int yorigin,
                                           int width, int height,
                                           int xdisplay, int ydisplay)
{
    /* Find the source position for the region. */

    int xsource = wrap_value(xorigin + xdisplay, s->image->width);
    int ysource = wrap_value(yorigin + ydisplay, s->image->height);

    /* Request tiling in the source coordinates. */

    image_tile_sprite(s, xsource, ysource,
                      width, height,
                      xdisplay, ydisplay);
}
